use crate::ast::named::Named;
use crate::ast::*;
use crate::name_analysis::fqn_context::FqnContext;
use crate::name_analysis::symbol::{FunctionSymbol, Symbol, VariableSymbol};
use crate::name_analysis::symbol_table::SymbolTable;
use crate::name_analysis::DiagnosticsContainer;
use codespan_reporting::diagnostic::{Diagnostic, Label};

pub(super) fn gather_module_level_declaration(
    declaration: &mut ModuleLevelDeclaration,
    symbol_table: &mut SymbolTable,
    fqn_context: &mut FqnContext,
    diagnostics: &mut DiagnosticsContainer,
) {
    use ModuleLevelDeclaration::*;
    match declaration {
        Function(function_definition) => {
            gather_function_definition(function_definition, symbol_table, fqn_context, diagnostics)
        }
        _ => todo!(),
    }
}

fn gather_function_definition(
    function: &mut FunctionDefinition,
    symbol_table: &mut SymbolTable,
    fqn_context: &mut FqnContext,
    diagnostics: &mut DiagnosticsContainer,
) {
    let declared_name = function.identifier.name().to_string();
    let resolved_name = fqn_context.resolve(&declared_name);

    let insert_result = symbol_table.insert(
        declared_name.clone(),
        Symbol::Function(FunctionSymbol {
            fqn: resolved_name.clone(),
            declared_name,
            is_public: function.is_public,
        }),
    );

    if let Err(msg) = insert_result {
        diagnostics.add(|file_id| {
            Diagnostic::error().with_label(Label::primary(file_id, function.identifier.range))
        });
    }

    function
        .identifier
        .set_scope_id(symbol_table.current_scope_id());

    symbol_table.push_scope(&format!("FunctionScope({})", resolved_name));
    gather_parameters(
        &mut function.parameters,
        symbol_table,
        fqn_context,
        diagnostics,
    );
    gather_function_body(&mut function.body, symbol_table, fqn_context, diagnostics);
    symbol_table.pop_scope();
}

fn gather_parameters(
    parameters: &mut Parameters,
    symbol_table: &mut SymbolTable,
    fqn_context: &mut FqnContext,
    diagnostics: &mut DiagnosticsContainer,
) {
    for parameter in &mut parameters.0 {
        gather_parameter(parameter, symbol_table, fqn_context, diagnostics);
    }
}

fn gather_parameter(
    parameter: &mut Parameter,
    symbol_table: &mut SymbolTable,
    fqn_context: &mut FqnContext,
    diagnostics: &mut DiagnosticsContainer,
) {
    let parameter_name = parameter.identifier.name();

    let insert_result = symbol_table.insert(
        parameter_name.to_string(),
        Symbol::Variable(VariableSymbol {
            name: parameter_name.to_string(),
            is_mutable: false,
        }),
    );
    
    if let Err(msg) = insert_result {
        diagnostics.add_error_at_identifier(&msg, &parameter.identifier);
    }
    
    parameter.identifier.set_scope_id(symbol_table.current_scope_id());
}

fn gather_function_body(
    function_body: &mut FunctionBody,
    symbol_table: &mut SymbolTable,
    fqn_context: &mut FqnContext,
    diagnostics: &mut DiagnosticsContainer,
) {
    use crate::ast::FunctionBody::*;
    match function_body {
        Block(block) => gather_block_statement_inner(block, symbol_table, fqn_context, diagnostics),
        _ => todo!(),
    }
}

fn gather_block_statement(
    block: &mut BlockStatement,
    symbol_table: &mut SymbolTable,
    fqn_context: &mut FqnContext,
    diagnostics: &mut DiagnosticsContainer,
) {
    symbol_table.push_scope("BlockStatementScope");
    gather_block_statement(block, symbol_table, fqn_context, diagnostics);
    symbol_table.pop_scope();
}

fn gather_block_statement_inner(
    block: &mut BlockStatement,
    symbol_table: &mut SymbolTable,
    fqn_context: &mut FqnContext,
    diagnostics: &mut DiagnosticsContainer,
) {
    for statement in &mut block.statements {
        gather_statement(statement, symbol_table, fqn_context, diagnostics);
    }
    if let Some(expression) = &mut block.expression {
        gather_expression(expression, symbol_table, diagnostics);
    }
}

fn gather_statement(
    statement: &mut Statement,
    symbol_table: &mut SymbolTable,
    fqn_context: &mut FqnContext,
    diagnostics: &mut DiagnosticsContainer,
) {
    use crate::ast::Statement::*;
    match statement {
        BlockStatement(block) => {
            gather_block_statement(block, symbol_table, fqn_context, diagnostics)
        }
        VariableDeclarationStatement(variable_declaration) => {
            gather_variable_declaration(variable_declaration, symbol_table, diagnostics)
        }
        AssignStatement(assign_statement) => {
            gather_assign_statement(assign_statement, symbol_table, fqn_context, diagnostics)
        }
        _ => todo!(),
    }
}

fn gather_variable_declaration(
    variable_declaration: &mut VariableDeclarationStatement,
    symbol_table: &mut SymbolTable,
    diagnostics: &mut DiagnosticsContainer,
) {
    let variable_name = variable_declaration.identifier.name().to_string();

    let insert_result = symbol_table.insert(
        variable_name.clone(),
        Symbol::Variable(VariableSymbol {
            name: variable_name,
            is_mutable: variable_declaration.is_mutable,
        }),
    );

    if let Err(msg) = insert_result {
        diagnostics.add_error_at_identifier(&msg, &variable_declaration.identifier);
    }

    variable_declaration
        .identifier
        .set_scope_id(symbol_table.current_scope_id());
    
    if let Some(initializer) = &mut variable_declaration.initializer {
        gather_expression(initializer, symbol_table, diagnostics);
    }
}

fn gather_assign_statement(
    assign_statement: &mut AssignStatement,
    symbol_table: &mut SymbolTable,
    fqn_context: &mut FqnContext,
    diagnostics: &mut DiagnosticsContainer,
) {
    gather_expression(&mut assign_statement.lhs, symbol_table, diagnostics);
    gather_expression(&mut assign_statement.rhs, symbol_table, diagnostics);
}

fn gather_expression(
    expression: &mut Expression,
    symbol_table: &mut SymbolTable,
    diagnostics: &mut DiagnosticsContainer,
) {
    use crate::ast::Expression::*;
    match expression {
        FullyQualifiedName(fully_qualified_name) => {
            gather_fully_qualified_name(fully_qualified_name, symbol_table);
        }
        _ => {}
    }
}

fn gather_fully_qualified_name(
    fully_qualified_name: &mut FullyQualifiedName,
    symbol_table: &mut SymbolTable,
) {
    fully_qualified_name.set_scope_id(symbol_table.current_scope_id());
}
