use crate::ast::named::Named;
use crate::ast::*;
use crate::name_analysis::symbol_table::SymbolTable;
use crate::name_analysis::DiagnosticsContainer;
use codespan_reporting::diagnostic::{Diagnostic, Label};

pub(super) fn resolve_module_level_declaration(
    declaration: &mut ModuleLevelDeclaration,
    symbol_table: &mut SymbolTable,
    diagnostics: &mut DiagnosticsContainer,
) {
    use crate::ast::ModuleLevelDeclaration::*;
    match declaration {
        Function(function_definition) => {
            resolve_function_definition(function_definition, symbol_table, diagnostics)
        }
        _ => todo!(),
    }
}

fn resolve_function_definition(
    function_definition: &mut FunctionDefinition,
    symbol_table: &mut SymbolTable,
    diagnostics: &mut DiagnosticsContainer,
) {
    resolve_function_body(&mut function_definition.body, symbol_table, diagnostics);
}

fn resolve_function_body(
    function_body: &mut FunctionBody,
    symbol_table: &mut SymbolTable,
    diagnostics: &mut DiagnosticsContainer,
) {
    use crate::ast::FunctionBody::*;
    match function_body {
        Block(block) => resolve_block(block, symbol_table, diagnostics),
        _ => todo!(),
    }
}

fn resolve_block(
    block_statement: &mut BlockStatement,
    symbol_table: &mut SymbolTable,
    diagnostics: &mut DiagnosticsContainer,
) {
    for statement in block_statement.statements.iter_mut() {
        resolve_statement(statement, symbol_table, diagnostics);
    }
    if let Some(expression) = block_statement.expression.as_mut() {
        resolve_expression(expression, symbol_table, diagnostics);
    }
}

fn resolve_statement(
    statement: &mut Statement,
    symbol_table: &mut SymbolTable,
    diagnostics: &mut DiagnosticsContainer,
) {
    use crate::ast::Statement::*;
    match statement {
        BlockStatement(block) => resolve_block(block, symbol_table, diagnostics),
        VariableDeclarationStatement(variable_declaration) => {
            resolve_variable_declaration(variable_declaration, symbol_table, diagnostics)
        }
        AssignStatement(assign_statement) => {
            resolve_assign_statement(assign_statement, symbol_table, diagnostics)
        }
        CallStatement(call_statement) => {
            resolve_call_statement(call_statement, symbol_table, diagnostics)
        }
        _ => todo!(),
    }
}

fn resolve_variable_declaration(
    variable_declaration: &mut VariableDeclarationStatement,
    symbol_table: &mut SymbolTable,
    diagnostics: &mut DiagnosticsContainer,
) {
    if let Some(initializer) = &mut variable_declaration.initializer {
        resolve_expression(initializer, symbol_table, diagnostics)
    }
}

fn resolve_assign_statement(
    assign_statement: &mut AssignStatement,
    symbol_table: &mut SymbolTable,
    diagnostics: &mut DiagnosticsContainer,
) {
    resolve_expression(&mut assign_statement.lhs, symbol_table, diagnostics);
    resolve_expression(&mut assign_statement.rhs, symbol_table, diagnostics);
}

fn resolve_call_statement(
    call_statement: &mut CallStatement,
    symbol_table: &mut SymbolTable,
    diagnostics: &mut DiagnosticsContainer,
) {
    resolve_expression(&mut call_statement.0, symbol_table, diagnostics)
}

fn resolve_expression(
    expression: &mut Expression,
    symbol_table: &mut SymbolTable,
    diagnostics: &mut DiagnosticsContainer,
) {
    use crate::ast::Expression::*;
    match expression {
        FullyQualifiedName(fqn) => resolve_fully_qualified_name(fqn, symbol_table, diagnostics),
        Literal(_) => {}
        _ => todo!(),
    }
}

fn resolve_fully_qualified_name(
    fully_qualified_name: &mut FullyQualifiedName,
    symbol_table: &mut SymbolTable,
    diagnostics: &mut DiagnosticsContainer,
) {
    let lookup_result = symbol_table.lookup(
        fully_qualified_name.name().as_ref(),
        fully_qualified_name.scope_id().expect(&format!(
            "FullyQualifiedName has no scope_id set: {:?}",
            fully_qualified_name
        )),
    );
    match lookup_result {
        Ok(symbol) => {
            fully_qualified_name.set_symbol(symbol.clone());
        }
        Err(e) => diagnostics.add(|file_id| {
            Diagnostic::error()
                .with_message(e)
                .with_label(Label::primary(file_id, fully_qualified_name.range))
        }),
    }
}
